package net.locusworks.argparser;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import net.locusworks.argparser.annotations.Parameter;
import net.locusworks.argparser.interfaces.IParameterConverter;

public class ArgParser {

  private Class<?> argClass;
  private List<Field> fields;
  private String programName;

  public ArgParser() { }

  public static ArgParser newBuilder() {
    return new ArgParser();
  }

  public ArgParser withArgClass(Class<?> argClass) {
    this.argClass = argClass;
    this.fields = Stream.of(argClass.getDeclaredFields())
        .filter(f -> f.isAnnotationPresent(Parameter.class))
        .map(f -> {
          f.setAccessible(true);
          return f;
        })
        .collect(Collectors.toList());
    
    if (fields == null || fields.isEmpty()) throw new IllegalArgumentException("No fields with @Parameter defined");
    
    fields.sort(new Comparator<Field>() {

      @Override
      public int compare(Field o1, Field o2) {
        Integer o1Order = o1.getDeclaredAnnotation(Parameter.class).order();
        Integer o2Order = o2.getDeclaredAnnotation(Parameter.class).order();
        return o1Order.compareTo(o2Order);
      }
    });
    
    boolean positional = true;
    for (Field f : fields) {
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      positional &= param.positional();
      if (!positional && param.positional()) throw new IllegalArgumentException("All positional values have to be first in list");
    }
    
    return this;
  }

  public ArgParser setProgramName(String name) {
    this.programName = name;
    return this;
  }

  public String help() {
    return "";
  }

  public <T> T parse(String[] args) {
    try {
      return parseItem(args);
    } catch (Exception ex) {
      System.err.println("An exception occured: " + ex.getMessage());
      System.err.println(help());
    }
    return null;
  }

  @SuppressWarnings("unchecked")
  private <T> T parseItem(String[] args) throws InstantiationException, IllegalAccessException {
    if (argClass == null) throw new IllegalArgumentException("Argument class was not specified");
    Object obj = argClass.newInstance();
    this.fields = Stream.of(obj.getClass().getDeclaredFields())
        .filter(f -> f.isAnnotationPresent(Parameter.class))
        .map(f -> {
          f.setAccessible(true);
          return f;
        })
        .collect(Collectors.toList());
    
    List<String> argList = new ArrayList<>(Arrays.asList(args));
    
    List<Field> positional = new ArrayList<>();
    for(Iterator<Field> iter = fields.iterator(); iter.hasNext();) {
      Field f = iter.next();
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      if (param.positional()) {
        positional.add(f);
        iter.remove();
      }
    }
    
    for(Iterator<String> iter = argList.iterator(); iter.hasNext() && !positional.isEmpty();) {
      String value = iter.next();
      Field f = positional.remove(0);
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      IParameterConverter<?> converter = param.converter().newInstance();
      f.set(obj, converter.convert(value));
      iter.remove();
    }
    
    Map<Field, List<String>> paramMap = new LinkedHashMap<>();
    List<String> params = null;
    for(Iterator<String> iter = argList.iterator(); iter.hasNext();) {
      String value = iter.next();
      Field f = findField(value);
      if (f != null) {
        if (!paramMap.containsKey(f)) {
          paramMap.put(f, new ArrayList<>());
        }
        params = paramMap.get(f);
      } else {
        params.add(value);
      }
      iter.remove();
    }
    
    for(Entry<Field, List<String>> entry : paramMap.entrySet()) {
      Field f = entry.getKey();
      List<String> values = entry.getValue();
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      if (param.variableArity() && !f.getType().isAssignableFrom(List.class))
        throw new IllegalArgumentException(String.format("Field %s is not of type list for the params", f.getName()));
      IParameterConverter<?> converter = param.converter().newInstance();
      List<Object> objs = values.stream().map(s -> converter.convert(s)).collect(Collectors.toList());
      if (param.variableArity()) {
        f.set(obj, getGenericList(f.getType(), objs));
      } else {
        f.set(obj, objs.get(0));
      }
    }
    
    return (T)obj;
  }

  private Field findField(String value) {
    for (Field f : fields) {
      Parameter param = f.getDeclaredAnnotation(Parameter.class);
      for (String s : Arrays.asList(param.names())) {
        if (s.equals(value)) return f;
      }
    }
    return null;
  }
  
  @SuppressWarnings("unchecked")
  private <Type> List<Type> getGenericList(Class<Type> type, List<Object> items) {
    List<Type> l = new ArrayList<>();
    for (Object o : items) {
      l.add((Type)o);
    }
    return l;
  }

}
